from __future__ import annotations
import inspect
import types
import typing
from PyQt5.QtCore import QDataStream

TEST_CONST = 123
class OpenTeleException(BaseException):
    """
    Base exception of the library.
    """

    def __init__(self, message : str = None, stack_index : int = 1) -> None:
        
        super().__init__(message if (message != None) else "")

        self.message = message
        self.desc = self.__class__.__name__

        stack = inspect.stack()

        self._caller_class = stack[stack_index][0].f_locals["self"].__class__ if "self" in stack[stack_index][0].f_locals else None
        self._caller_method = stack[stack_index][0].f_code.co_name
        self.stack = stack

        if self._caller_method != "<module>":
            args,_,_,locals = inspect.getargvalues(self.stack[stack_index][0])
            parameters = { arg : locals[arg] for arg in args}
            self._caller_method_params = "".join(f'{i}={parameters[i]}, ' for i in parameters)[:-2]
        else:
            self._caller_method = "__main__"
            self._caller_method_params = ""


        if self.desc == "OpenTeleException":
            self.desc = "Unexpected Exception"


    def __str__(self):
        reason = self.desc.__str__()

        if self.message != None:
            reason += f": {self.message}"

        reason += " [ Called by "
        if self._caller_class != None:
            
            parent_list = []
            base = self._caller_class
            while hasattr(base, "__name__"):
                parent_list.append(base.__name__)
                base = base.__base__

            parent_list.reverse()
            reason += "".join(f"{i}." for i in parent_list[1:])
            reason += self._caller_method + "() ]"
            # reason += f"\n>\t{self._caller_method}({self._caller_method_params})"
        else:
            reason += f"{self._caller_method}() ]"
            # reason += f"{self._caller_method}({self._caller_method_params}) ]"
        return reason
        
class TFileNotFound(OpenTeleException):
    '''
    Could not find or open the file
    '''

class TDataInvalidMagic(OpenTeleException):
    '''
    TData file has an invalid magic data, which is the first 4 bytes of the file\n
    This usually mean that the file is corrupted or not in the supported formats
    '''

class TDataInvalidCheckSum(OpenTeleException):
    '''
    TData file has an invalid checksum\n
    This usually mean that the file is corrupted or not in the supported formats
    '''

class TDataBadDecryptKey(OpenTeleException):
    '''
    Could not decrypt the data with this key\n
    This usually mean that the file is password-encrypted
    '''

class TDataWrongPasscode(OpenTeleException):
    '''
    Wrong passcode to decrypt tdata folder\n
    '''

class TDataBadEncryptedDataSize(OpenTeleException):
    '''
    The encrypted data size part of the file is corrupted
    '''
class TDataBadDecryptedDataSize(OpenTeleException):
    '''
    The decrypted data size part of the file is corrupted
    '''
class TDataBadConfigData(OpenTeleException):
    '''
    TData contains bad config data that couldn't be parsed
    '''

class QDataStreamFailed(OpenTeleException):
    '''
    Could not stream data from QDataStream\n
    Please check the QDataStream.status() for more information
    '''

class AccountAuthKeyNotFound(OpenTeleException):
    '''
    Account.authKey is missing, something went wrong
    '''

class TDataReadMapDataFailed(OpenTeleException):
    '''
    Could not read map data
    '''
class TDataReadMapDataIncorrectPasscode(OpenTeleException):
    '''
    Could not read map data because of incorrect passcode
    '''
class TDataAuthKeyNotFound(OpenTeleException):
    '''
    Could not find authKey in TData
    '''
class MaxAccountLimit(OpenTeleException):
    '''
    Maxed out limit for accounts per tdesktop client
    '''

class TDesktopUnauthorized(OpenTeleException):
    '''
    TDesktop client is unauthorized
    '''

class TelethonUnauthorized(OpenTeleException):
    '''
    Telethon client is unauthorized
    '''

class TDataSaveFailed(OpenTeleException):
    '''
    Could not save TDesktop to tdata folder
    '''

class TDesktopNotLoaded(OpenTeleException):
    '''
    TDesktop instance has no account
    '''
class TDesktopHasNoAccount(OpenTeleException):
    '''
    TDesktop instance has no account
    '''

class TDAccountNotLoaded(OpenTeleException):
    '''
    TDesktop account hasn't been loaded yet
    '''
class NoPasswordProvided(OpenTeleException):
    '''
    You can't live without a password bro
    '''

class PasswordIncorrect(OpenTeleException):
    '''
    incorrect passwrd
    '''

class LoginFlagInvalid(OpenTeleException):
    '''
    Invalid login flag
    '''

class NoInstanceMatched(OpenTeleException):
    '''
    Invalid login flag
    '''

@typing.overload
def Expects(condition : bool, 
            message : str = None,
            done : typing.Callable[[],None] = None, 
            fail : typing.Callable[[OpenTeleException],None] = None, 
            silent : bool = False,
            stack_index : int = 1) -> bool:
    """Expect a condition to be `True`, raise an `OpenTeleException` if it's not.
    
    ### Arguments:
        condition (bool):
            Condition that you're expecting.

        message (str, default=None):
            Custom exception message
    
        done (`lambda`, default=None):
            lambda to execute when done without error
    
        fail (`lambda`, default=None):
            lambda to execute when the condition is False, the lambda will be execute before raising the exception.
    
        silent (`bool`, default=False):
            if True then it won't raise the exception, only execute fail lambda.
    
        stack_index (`int`, default=1):
            stack index to raise the exception with trace back to where it happens, intended for internal usage.
    
    ### Raises:
        `OpenTeleException`: exception
    """
    pass


@typing.overload
def Expects(condition : bool, 
            exception : OpenTeleException = None, 
            done : typing.Callable[[],None] = None, 
            fail : typing.Callable[[OpenTeleException],None] = None, 
            silent : bool = False,
            stack_index : int = 1) -> bool:
            
    """Expect a condition to be `True`, raise an `OpenTeleException` if it's not.
    
    ### Arguments:
        condition (bool):
            Condition that you're expecting.

        message (OpenTeleException, default=None):
            Custom exception.
    
        done (`lambda`, default=None):
            lambda to execute when done without error.
    
        fail (`lambda`, default=None):
            lambda to execute when the condition is False, the lambda will be execute before raising the exception.
    
        silent (`bool`, default=False):
            if True then it won't raise the exception, only execute fail lambda.
    
        stack_index (`int`, default=1):
            stack index to raise the exception with trace back to where it happens, intended for internal usage.
    
    ### Raises:
        `OpenTeleException`: exception
    """
    pass

def Expects(condition : bool, 
            exception : typing.Union[OpenTeleException, str] = None, 
            done : typing.Callable[[],None] = None, 
            fail : typing.Callable[[OpenTeleException],None] = None, 
            silent : bool = False,
            stack_index : int = 1) -> bool:
    
    if condition:
        if done != None: done()
        return condition
    
    if isinstance(exception, str): 
        exception = OpenTeleException(exception, 2)
    elif exception == None or isinstance(exception, OpenTeleException): 
        pass
    else:
        raise OpenTeleException("No instance of Expects() match the arguments given", 2)


    if exception == None:
        exception =  OpenTeleException("Unexpected error", 2)
        
    
    # no raise exception
    if silent:
        if fail != None: fail(exception)
        return condition

    else:
        stack = inspect.stack()
        frame = stack[stack_index].frame
        tb = types.TracebackType(None, frame, frame.f_lasti, frame.f_lineno) # type: ignore
        exception = exception.with_traceback(tb)

        if fail != None: fail(exception)

        raise exception

def ExpectStreamStatus(stream : QDataStream, message : str = "Could not stream data"):
    Expects(stream.status() == QDataStream.Status.Ok,
            stack_index=2,
            exception=QDataStreamFailed("Could not read keys count from mtp authorization.", stack_index=2))



import inspect
import os

from typing import Any

IS_DEBUG_MODE = False

if IS_DEBUG_MODE:
    from rich import print

class HookClassMethod(type):

    def __get__(self, obj, cls):
        self.__owner__ = obj if obj else cls
        return self

    def __call__(self, *args, **kwargs) -> Any:
        
        stack = inspect.stack()[1]
        
        location = f"{os.path.basename(stack.filename)[:-3]}"
        lineno = f"{stack.lineno}."

        if len(location) >= 15:
            location = ".." + location[-15:]
        
        location = location.ljust(15) + "| " + lineno.ljust(6)

        context = stack.code_context[0][:-1].strip() # type: ignore

        print(location + context)

        return self.__fget__(self.__owner__, *args, **kwargs) # type: ignore

    def __set_name__(self, owner, name):
        self.__owner__ = owner
        self.__ownername__ = owner.__name__
        self.__fname__ = name
        pass
        
    def __new__(cls, decorated_func):

        firstdct = dict(decorated_func.__dict__)
        for i, x in cls.__dict__.items():
            firstdct [i] = x


        result = type.__new__(cls, decorated_func.__name__, decorated_func.__class__.__bases__, firstdct)
        result.__fget__ = decorated_func # type: ignore
        return result

def parse_arg(value) -> str:
    if isinstance(value, type):
        return value.__name__
    elif isinstance(value, str):
        return f"'{value}'"
    elif isinstance(value, int):
        return f"{value}"
    elif isinstance(value, object):
        return value.__class__.__name__ + "(...)"
    return value



from __future__ import annotations


from .configs import *
from . import shared as td

import logging
from telethon.network.connection.connection import Connection
from telethon.network.connection.tcpfull import ConnectionTcpFull
from telethon.sessions.abstract import Session

# if TYPE_CHECKING:
#     from . import *


class TDesktop(BaseObject):
    """
    Telegram Desktop client.
    
    A client can have multiple accounts, up to 3 - according to official Telegram Desktop client.

    ### Attributes:
        api (`API`):
            The API this client is using.
        
        accountsCount (`int`):
            The numbers of accounts in this client.

        accounts (`List[Account]`):
            List of accounts in this client.

        mainAccount (`Account`):
            The main account of this client.
        
        basePath (`str`):
            The path to `tdata folder`.

        passcode (`str`):
            Passcode of the client, the same as Local Passcode on `Telegram Desktop`.\\
            Use to encrypt and decrypt `tdata files`.

        AppVersion (`int`):
            App version of the client.
        
        kMaxAccounts (`int`):
            See `kMaxAccounts`.

        keyFile (`str`):
            See `keyFile`.
        
        kDefaultKeyFile (`str`):
            Default value for `keyFile`.
    
    ### Methods:
        LoadTData():
            Load the client from `tdata folder`. \\
            Use this if you didn't set the `basePath` when initializing the client.

        SaveTData():
            Save the client session to `tdata folder` - which can be used by `Telegram Desktop`.

        isLoaded():
            Return `True` if the client has successfully loaded accounts from `tdata` or `TelegramClient`.

        ToTelethon():
            Convert this session to `TelegramClient`.

        FromTelethon():
            Create a new session from `TelegramClient`.

    """
    
    @typing.overload
    def __init__(self) -> None:
        pass

    @typing.overload
    def __init__(self,  basePath : str = None,
                        api : Union[Type[APIData], APIData] = API.TelegramDesktop) -> None:
        pass
                        
    @typing.overload
    def __init__(self,  basePath : str = None,
                        api : Union[Type[APIData], APIData] = API.TelegramDesktop,
                        passcode : str = None,
                        keyFile : str = None) -> None:
        pass

    def __init__(self,  basePath : str = None,
                        api : Union[Type[APIData], APIData] = API.TelegramDesktop,
                        passcode : str = None,
                        keyFile : str = None) -> None:
        """
        Initialize a `TDesktop` client

        ### Arguments:
            basePath (`str`, default=`None`):
                The path to the `tdata folder`.
                If the path doesn't exists or its data is corrupted, a new instance will be created.
            
            api (`API`, default=`TelegramDesktop`):
                Which API to use. Read more `[here](API)`.

            passcode (str, default=`None`):
                The passcode for tdata, same as the Local Passcode on `Telegram Desktop`.
        
            keyFile (str, default="data"):
                See `keyFile`.
        """
        self.__accounts : typing.List[td.Account] = []
        self.__basePath = basePath
        self.__keyFile = keyFile if (keyFile != None) else TDesktop.kDefaultKeyFile
        self.__passcode = passcode if (passcode != None) else str("")
        self.__passcodeBytes = self.__passcode.encode("ascii")
        self.__mainAccount : Optional[td.Account] = None
        self.__active_index = -1
        self.__passcodeKey = None
        self.__localKey = None
        self.__AppVersion = None
        self.__isLoaded = False
        self.__api  = api.copy()

        if basePath != None:
            self.__basePath = td.Storage.GetAbsolutePath(basePath)
            self.LoadTData()
 

    def isLoaded(self) -> bool:
        """
        Return `True` if the client has successfully loaded accounts from `tdata` or `TelegramClient`
        """
        return self.__isLoaded

    def LoadTData(self, basePath : str = None, passcode : str = None, keyFile : str = None):
        """
        Loads accounts from `tdata folder`

        ### Arguments:
            basePath (`str`, default=`None`):
                The path to the folder.
        
            passcode (`str`, default=`None`):
                Read more `[here](passcode)`.
        
            keyFile (`str`, default=`None`):
                Read more `[here](keyFile)`.
        
        ### Raises:
            `TDataBadDecryptKey`: The `tdata folder` is password-encrypted, please the set the argument `passcode` to decrypt it.

        ### Warning:
            This function is not recommended to use:
                You should load tdata using `TDesktop(basePath="path")`.
                Don't manually load tdata using this function, bugs might pop up out of nowhere.

        ### Examples:
        ```python
            # Using the API that we've generated before. Please refer to method API.Generate() to learn more.
            oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old.session")
            oldclient = TelegramClient("old.session", api=oldAPI)
            await oldClient.connect()

            # We can safely use CreateNewSession with a different API.
            # Be aware that you should not use UseCurrentSession with a different API than the one that first authorized it.
            newAPI = API.TelegramAndroid.Generate("new_tdata")
            tdesk = await TDesktop.FromTelethon(oldclient, flag=CreateNewSession, api=newAPI)

            # Save the new session to a folder named "new_tdata"
            tdesk.SaveTData("new_tdata")
        ```
        """
        
        if basePath == None: basePath = self.basePath
        
        Expects(basePath != None and basePath != "", "No folder provided to load tdata")
        
        if keyFile != None and self.__keyFile != keyFile:
            self.__keyFile = keyFile

        if passcode != None and self.__passcode != passcode:
            self.__passcode = passcode
            self.__passcodeBytes = passcode.encode("ascii")
        
        try:
            self.__loadFromTData()
            
        except OpenTeleException as e:
            if isinstance(e, TDataBadDecryptKey):
                if self.passcode == "":
                    raise TDataBadDecryptKey("The tdata folder is password-encrypted, please the set the argument 'passcode' to decrypt it")
                else:
                    raise TDataBadDecryptKey("Failed to decrypt tdata folder because of invalid passcode")
            else:
                raise e
        
        Expects(self.isLoaded(), "Failed to load? Something went seriously wrong")


    def SaveTData(self, basePath : str = None, passcode : str = None, keyFile : str = None) -> bool:
        """
        Save the client session to a folder.

        ### Arguments:
            basePath (`str`, default=None):
                Path to the folder\\
                If None then the data will be saved at the basePath given at creation.

            passcode (`str`, default=`None`):
                Read more `[here](passcode)`.
        
            keyFile (`str`, default=`None`):
                Read more `[here](keyFile)`.

        ### Examples:
            Save a telethon session to tdata:
        ```python
            # Using the API that we've generated before. Please refer to method API.Generate() to learn more.
            oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old.session")
            oldclient = TelegramClient("old.session", api=oldAPI)
            await oldClient.connect()

            # We can safely CreateNewSession with a different API.
            # Be aware that you should not use UseCurrentSession with a different API than the one that first authorized it.
            newAPI = API.TelegramAndroid.Generate("new_tdata")
            tdesk = await TDesktop.FromTelethon(oldclient, flag=CreateNewSession, api=newAPI)

            # Save the new session to a folder named "new_tdata"
            tdesk.SaveTData("new_tdata")
        ```
        """
        if basePath == None: basePath = self.basePath
        
        self.__keyFile = keyFile if (keyFile != None and self.keyFile != keyFile) else self.keyFile

        Expects(basePath != None and basePath != "", "No folder provided to save tdata")
    
        if passcode != None and self.__passcode != passcode:
            self.__passcode = passcode
            self.__passcodeBytes = passcode.encode("ascii")
            self.__isLoaded = False # to generate new localKey

        if not self.isLoaded(): self.__generateLocalKey()
        Expects(self.isLoaded(), "Failed to load? Something went seriously wrong")

        try:

            basePath = td.Storage.GetAbsolutePath(basePath)
            if not self.basePath: self.__basePath = basePath

            self.__writeAccounts(basePath, keyFile)
            return True

        except OpenTeleException as e:
            raise TDataSaveFailed("Could not save tdata, something went wrong") from e

    def __writeAccounts(self, basePath : str, keyFile : str = None) -> None:
        # Intended for internal usage only

        Expects(len(self.accounts) > 0)
        Expects(basePath != None and basePath != "", "No folder provided to save tdata")

        for account in self.accounts:
            account._writeData(basePath, keyFile)

        key = td.Storage.FileWriteDescriptor("key_" + self.keyFile, basePath)
        key.writeData(self.__passcodeKeySalt)        
        key.writeData(self.__passcodeKeyEncrypted)

        keySize = sizeof(int32) + sizeof(int32) * len(self.accounts)
        keyData = td.Storage.EncryptedDescriptor(keySize)
        keyData.stream.writeInt32(len(self.accounts))

        for account in self.accounts:
            keyData.stream.writeInt32(account.index)

        keyData.stream.writeInt32(self.__active_index)
        key.writeEncrypted(keyData, self.__localKey) # type: ignore
        key.finish()

    def __generateLocalKey(self) -> None:
        # Intended for internal usage only

        Expects(not self.isLoaded())

        LocalEncryptSaltSize = 32

        _pass = td.Storage.RandomGenerate(td.AuthKey.kSize)
        _salt = td.Storage.RandomGenerate(LocalEncryptSaltSize)
        self.__localKey = td.Storage.CreateLocalKey(_salt, _pass)

        self.__passcodeKeySalt = td.Storage.RandomGenerate(LocalEncryptSaltSize) # LocalEncryptSaltSize = 32
        self.__passcodeKey = td.Storage.CreateLocalKey(self.__passcodeKeySalt, QByteArray(self.__passcodeBytes))

        passKeyData = td.Storage.EncryptedDescriptor(td.AuthKey.kSize)
        self.__localKey.write(passKeyData.stream)

        self.__passcodeKeyEncrypted = td.Storage.PrepareEncrypted(passKeyData, self.__passcodeKey)

        # set new local key for self.accounts
        for account in self.accounts:
            account.localKey = self.localKey

        self.__isLoaded = True

    def _addSingleAccount(self, account : td.Account):
        # Intended for internal usage only

        # Expects(self.isLoaded(), "Could not add account because i haven't been loaded")
        Expects(account.isLoaded(), "Could not add account because the account hasn't been loaded")
        
        account.localKey = self.localKey
            
        self.__accounts.append(account)
    
        if self.mainAccount == None:
            self.__mainAccount = self.__accounts[0]

    def __loadFromTData(self) -> None:
        # Intended for internal usage only

        Expects(self.basePath != None and self.basePath != "", "No folder provided to load tdata")

        self.accounts.clear()

        # READ KEY_DATA
        keyData = td.Storage.ReadFile("key_" + self.keyFile, self.basePath) # type: ignore

        salt, keyEncrypted, infoEncrypted = QByteArray(), QByteArray(), QByteArray()

        keyData.stream >> salt >> keyEncrypted >> infoEncrypted

        Expects(keyData.stream.status() == QDataStream.Status.Ok, 
                QDataStreamFailed("Failed to stream keyData"))

        
        self.__AppVersion = keyData.version
        self.__passcodeKey = td.Storage.CreateLocalKey(salt, QByteArray(self.__passcodeBytes))

        keyInnerData = td.Storage.DecryptLocal(keyEncrypted, self.passcodeKey) # type: ignore

        self.__localKey = td.AuthKey(keyInnerData.stream.readRawData(256))
        self.__passcodeKeyEncrypted = keyEncrypted
        self.__passcodeKeySalt = salt

        info = td.Storage.DecryptLocal(infoEncrypted, self.localKey) # type: ignore
        count = info.stream.readInt32()

        Expects(count > 0, "accountsCount is zero, the data might has been corrupted")

        for i in range(count):
            index = info.stream.readInt32()
            if (index >= 0) and (index < TDesktop.kMaxAccounts):
                try:
                    account = td.Account(self, basePath=self.basePath, api=self.api, keyFile=self.keyFile, index=index)
                    account.prepareToStart(self.__localKey)

                    if account.isLoaded():
                        self.accounts.append(account)
                        
                except OpenTeleException as e:
                    pass
        
        Expects(len(self.accounts) > 0, "No account has been loaded")
        
        self.__active_index = 0
        if (not info.stream.atEnd()):
            self.__active_index = info.stream.readInt32()
        
        for account in self.accounts:
            if (account.index == self.__active_index):
                self.__mainAccount = account
                break

        if not self.__mainAccount:
            self.__mainAccount = self.accounts[0]
        
        self.__isLoaded = True


# EXTENDED FUNCTION ====================================================================
# @extend_classs
# class TDesktop:

    @typing.overload
    async def ToTelethon(self,
                    session         : Union[str, Session] = None,
                    flag            : Type[LoginFlag] = CreateNewSession,
                    api             : Union[Type[APIData], APIData] = API.TelegramDesktop,
                    password        : str = None) -> tl.TelegramClient:
        """
        
        ### Arguments:
            session (`str`, `Session`, default=`None`):
                The file name of the `session file` to be used, if `None` then the session will not be saved.\\
                Read more [here](https://docs.telethon.dev/en/latest/concepts/sessions.html?highlight=session#what-are-sessions).
        
            flag (`LoginFlag`, default=`CreateNewSession`):
                The login flag. Read more `[here](LoginFlag)`.
        
            api (`APIData`, default=`TelegramDesktop`):
                Which API to use. Read more `[here](API)`.
        
            password (`str`, default=`None`):
                Two-step verification password if needed.
        
        ### Returns:
            - Return an instance of `TelegramClient` on success

        ### Examples:
            Create a telethon session from tdata folder:
        ```python
            # Using the API that we've generated before. Please refer to method API.Generate() to learn more.
            oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old_tdata")
            tdesk = TDesktop("old_tdata", api=oldAPI)

            # We can safely authorize the new client with a different API.
            newAPI = API.TelegramAndroid.Generate(unique_id="new.session")
            client = await tdesk.ToTelethon(session="new.session", flag=CreateNewSession, api=newAPI)
            await client.connect()
            await client.PrintSessions()
        ```
        """

    @typing.overload
    async def ToTelethon( self,
                    session         : Union[str, Session] = None,
                    flag            : Type[LoginFlag] = CreateNewSession,
                    api             : Union[Type[APIData], APIData] = API.TelegramDesktop,
                    password                : str = None,
                    *,
                    connection              : typing.Type[Connection] = ConnectionTcpFull,
                    use_ipv6                : bool = False,
                    proxy                   : Union[tuple, dict] = None,
                    local_addr              : Union[str, tuple] = None,
                    timeout                 : int = 10,
                    request_retries         : int = 5,
                    connection_retries      : int = 5,
                    retry_delay             : int = 1,
                    auto_reconnect          : bool = True,
                    sequential_updates      : bool = False,
                    flood_sleep_threshold   : int = 60,
                    raise_last_call_error   : bool = False,
                    loop                    : asyncio.AbstractEventLoop = None,
                    base_logger             : Union[str, logging.Logger] = None,
                    receive_updates         : bool = True) -> tl.TelegramClient:
        pass

    async def ToTelethon( self,
                    session         : Union[str, Session] = None,
                    flag            : Type[LoginFlag] = CreateNewSession,
                    api             : Union[Type[APIData], APIData] = API.TelegramDesktop,
                    password        : str = None,
                    *,
                    connection              : typing.Type[Connection] = ConnectionTcpFull,
                    use_ipv6                : bool = False,
                    proxy                   : Union[tuple, dict] = None,
                    local_addr              : Union[str, tuple] = None,
                    timeout                 : int = 10,
                    request_retries         : int = 5,
                    connection_retries      : int = 5,
                    retry_delay             : int = 1,
                    auto_reconnect          : bool = True,
                    sequential_updates      : bool = False,
                    flood_sleep_threshold   : int = 60,
                    raise_last_call_error   : bool = False,
                    loop                    : asyncio.AbstractEventLoop = None,
                    base_logger             : Union[str, logging.Logger] = None,
                    receive_updates         : bool = True) -> tl.TelegramClient:

        Expects(self.isLoaded(), TDesktopNotLoaded("You need to load accounts from a tdata folder first"))
        Expects(self.accountsCount > 0, TDesktopHasNoAccount("There is no account in this instance of TDesktop"))
        assert self.mainAccount
        
        return await tl.TelegramClient.FromTDesktop(self.mainAccount, session=session, flag=flag, api=api, password=password,
                                            connection=connection, use_ipv6=use_ipv6,
                                            proxy=proxy, local_addr=local_addr, timeout=timeout, request_retries=request_retries,
                                            connection_retries=connection_retries, retry_delay=retry_delay, auto_reconnect=auto_reconnect,
                                            sequential_updates=sequential_updates, flood_sleep_threshold=flood_sleep_threshold,
                                            raise_last_call_error=raise_last_call_error, loop=loop, base_logger=base_logger,
                                            receive_updates=receive_updates)

    @staticmethod
    async def FromTelethon(telethonClient : tl.TelegramClient,
                    flag            : Type[LoginFlag] = CreateNewSession,
                    api             : Union[Type[APIData], APIData] = API.TelegramDesktop,
                    password        : str = None) -> TDesktop:
        """
        Create an instance of `TDesktop` from `TelegramClient`.

        ### Arguments:
            telethonClient (`TelegramClient`):
                The `TelegramClient` you want to convert from.
        
            flag (`LoginFlag`, default=`CreateNewSession`):
                The login flag. Read more `[here](LoginFlag)`.
        
            api (`APIData`, default=`API.TelegramDesktop`):
                Which API to use. Read more `[here](API)`.
        
            password (`str`, default=`None`):
                Two-step verification password if needed.

        ### Examples:
            Save a telethon session to tdata:
        ```python
            # Using the API that we've generated before. Please refer to method API.Generate() to learn more.
            oldAPI = API.TelegramDesktop.Generate(system="windows", unique_id="old.session")
            oldclient = TelegramClient("old.session", api=oldAPI)
            await oldClient.connect()

            # We can safely CreateNewSession with a different API.
            # Be aware that you should not use UseCurrentSession with a different API than the one that first authorized it.
            newAPI = API.TelegramAndroid.Generate("new_tdata")
            tdesk = await TDesktop.FromTelethon(oldclient, flag=CreateNewSession, api=newAPI)

            # Save the new session to a folder named "new_tdata"
            tdesk.SaveTData("new_tdata")
        ```
        """
        
        Expects((flag == CreateNewSession) or (flag == UseCurrentSession), LoginFlagInvalid("LoginFlag invalid"))
        
        _self = TDesktop()
        _self.__isLoaded = True
        
        await td.Account.FromTelethon(telethonClient, flag=flag, api=api, password=password, owner=_self)

        return _self
         
    kMaxAccounts : int = int(3)
    """The maximum amount of accounts a client can have"""

    kDefaultKeyFile = "data"
    """See `TDesktop.keyFile`"""

    @property
    def api(self) -> APIData:
        """
        The API this client is using.
        """
        return self.__api

    @api.setter
    def api(self, value) -> None:
        self.__api = value
        for account in self.accounts:
            account.api = value
    
    @property
    def basePath(self) -> Optional[str]:
        """
        Base folder of `TDesktop`, this is where data stored
        Same as tdata folder of `Telegram Desktop`
        """
        return self.__basePath

    @property
    def passcode(self) -> str:
        """
        Passcode
        Passcode used to encrypt and decrypt data
        Same as the Local Passcode of `Telegram Desktop`
        """
        return self.__passcode

    @property
    def keyFile(self) -> str:
        """
        The default value is `"data"`, this argument is rarely ever used.
        It is used by `Telegram Desktop` by running it with the `"-key"` argument.
        I don't know what's the use cases of it, maybe this was a legacy feature of `Telegram Desktop`.
        """
        return self.__keyFile

    @property
    def passcodeKey(self) -> Optional[td.AuthKey]:
        return self.__passcodeKey

    @property
    def localKey(self) -> Optional[td.AuthKey]:
        """
        The key used to encrypt/decrypt data
        """
        return self.__localKey

    @property
    def AppVersion(self) -> Optional[int]:
        """
        App version of TDesktop client
        """
        return self.__AppVersion

    @property
    def AppVersionString(self) -> Optional[str]:
        raise NotImplementedError()
        return self.__AppVersion

    @property
    def accountsCount(self) -> int:
        """
        The number of accounts this client has
        """
        # return self.__accountsCount
        return len(self.__accounts)

    @property
    def accounts(self) -> List[td.Account]:
        """
        List of accounts this client has\n
        If you want to get the main account, please use .mainAccount instead
        """
        return self.__accounts

    @property
    def mainAccount(self) -> Optional[td.Account]:
        """
        The main account of the client
        """
        return self.__mainAccount


from __future__ import annotations

from . import debug

from typing import Coroutine, Tuple, Type, Callable, TypeVar, Optional, List, Any, Dict
from types import FunctionType

import abc

APP_VERSION = 3004000
TDF_MAGIC = b"TDF$"

_T = TypeVar('_T')
_TCLS = TypeVar('_TCLS', bound=type)
_RT = TypeVar('_RT')
_F = TypeVar('_F', bound=Callable[..., Any])


class BaseMetaClass(abc.ABCMeta):
    
    def __new__(cls : Type[_T], clsName : str, bases : Tuple[type], attrs : Dict[str, Any]) -> _T:

        # Hook all subclass methods
        if debug.IS_DEBUG_MODE:
            ignore_list = ["__new__", "__get__", "__call__", "__set_name__", "__str__", "__repr__"]

            for attr, val in attrs.items():
                if not attr in ignore_list and callable(val) and not isinstance(val, type):
                    newVal = debug.HookClassMethod(val)
                    attrs[attr] = newVal

        return super().__new__(cls, clsName, bases, attrs)


class BaseObject(object, metaclass=BaseMetaClass):
    pass


class override(object):
    """
    To use inside a class decorated with @extend_class\n
    Any attributes decorated with @override will be replaced
    """
    def __new__(cls, decorated_func : _F) -> _F:
        
        # check if decorated_cls really is a function
        if not isinstance(decorated_func, FunctionType):
            raise BaseException("@override decorator is only for functions, not classes")

        decorated_func.__isOverride__ = True # type: ignore
        return decorated_func # type: ignore

    @staticmethod
    def isOverride(func : _F) ->bool:
        if not hasattr(func, "__isOverride__"): return False
        return func.__isOverride__

class extend_class(object):
    """
    Extend a class, all attributes will be added to its parents\n
    This won't override attributes that are already existed, please refer to @override or @extend_override_class to do this
    """
    def __new__(cls, decorated_cls : _TCLS, isOverride : bool = False) -> _TCLS:
        
        # check if decorated_cls really is a class (type)
        if not isinstance(cls, type):
            raise BaseException("@extend_class decorator is only for classes, not functions")

        newAttributes = dict(decorated_cls.__dict__)
        crossDelete = ["__abstractmethods__", "__module__", "_abc_impl", "__doc__"]
        [(newAttributes.pop(cross) if cross in newAttributes else None) for cross in crossDelete]

        crossDelete = {}
        
        base = decorated_cls.__bases__[0]

        if not isOverride:
            # loop through its parents and add attributes

            for attributeName, attributeValue in newAttributes.items():
                
                # check if class base already has this attribute
                result = extend_class.getattr(base, attributeName)

                if result != None:
                    if id(result['value']) == id(attributeValue):
                        crossDelete[attributeName] = attributeValue
                    else:
                    
                        # if not override this attribute
                        if not override.isOverride(attributeValue):
                            print(f"[{attributeName}] {id(result['value'])} - {id(attributeValue)}")
                            raise BaseException("err")
            
            [newAttributes.pop(cross) for cross in crossDelete]
            
    
        for attributeName, attributeValue in newAttributes.items():
            
            # let's backup this attribute for future uses
            result = extend_class.getattr(base, attributeName)

            if result != None:
                # ! dirty code, gonna fix it later, it's okay for now
                setattr(base, f"__{decorated_cls.__name__}__{attributeName}", result['value'])
                setattr(decorated_cls, f"__{decorated_cls.__name__}__{attributeName}", result['value'])

            setattr(base, attributeName, attributeValue)

        return decorated_cls
    
    @staticmethod
    def object_hierarchy_getattr(obj : object, attributeName : str) -> List[str]:
        
        results = []
        if type(obj) == object: return results

        if attributeName in obj.__dict__:
            val = obj.__dict__[attributeName]
            results.append({"owner": obj, "value": val})

        if attributeName in obj.__class__.__dict__:
            val = obj.__class__.__dict__[attributeName]
            results.append({"owner": obj, "value": val})

        for base in obj.__bases__: # type: ignore
            results += extend_class.object_hierarchy_getattr(base, attributeName)


        results.reverse()
        return results

    @staticmethod
    def getattr(obj : object, attributeName : str) -> Optional[dict]:
        try:
            value = getattr(obj, attributeName)
            return {"owner" : obj, "value" : value}
        except BaseException as e:
            return None
            
class extend_override_class(extend_class):
    """
    Extend a class, all attributes will be added to its parents\n
    If those attributes are already existed, they will be replaced by the new one
    """
    def __new__(cls, decorated_cls : _TCLS) -> _TCLS:
        return super().__new__(cls, decorated_cls, True)

class sharemethod(object):

    def __get__(self, obj, cls):
        self.__owner__ = obj if obj else cls
        return self

    def __call__(self, *args) -> Any:
        return self.__fget__.__get__(self.__owner__)(*args) # type: ignore

    def __set_name__(self, owner, name):
        self.__owner__ = owner
        pass
        
    def __new__(cls, func : _F) -> _F:
        result = super().__new__(cls)
        result.__fget__ = func # type: ignore
        return result # type: ignore





from __future__ import annotations

from enum import IntEnum

from ..exception import *
from ..utils import *
from ..api import *
from .. import tl

from typing import Literal, Union, Callable, TypeVar, Type, Optional, List, Dict, Any, TYPE_CHECKING
from ctypes import sizeof, c_int32 as int32, c_int64 as int64, c_uint32 as uint32, c_uint64 as uint64, c_short as short, c_ushort as ushort
from PyQt5.QtCore import QByteArray, QDataStream, QBuffer, QIODevice, QSysInfo, QDir, QFile
from types import FunctionType

import asyncio
    

APP_VERSION = 3004000
TDF_MAGIC = b"TDF$"


_T = TypeVar('_T')
_T2 = TypeVar('_T2')
_TCLS = TypeVar('_TCLS', bound=type)
_RT = TypeVar('_RT')

_F = TypeVar('_F', bound=Callable[..., Any])


class BareId(int):
    '''
    BareId
    '''

class ChatIdType(BaseObject):
    '''
    ChatIdType
    '''
    bare = BareId(0)
    kShift = BareId(0)
    kReservedBit = BareId(0x80)

    def __init__(self, value : BareId) -> None:
        self.bare = value

class UserId(ChatIdType):
    kShift = BareId(0)

class ChatId(ChatIdType):
    kShift = BareId(1)

class ChannelId(ChatIdType):
    kShift = BareId(2)

class FakeChatId(ChatIdType):
    kShift = BareId(0x7F)

class PeerId(int):
    '''
    PeerId
    '''
    kChatTypeMask = BareId(0xFFFFFFFFFFFF)
    def __init__(self, value) -> None:
        self.value = value
    
    def Serialize(self):
        Expects(not (self.value & (UserId.kReservedBit << 48)))
        return self.value | (UserId.kReservedBit << 48)

    @staticmethod
    def FromChatIdType(id : typing.Union[UserId, ChatId, ChannelId, FakeChatId]) -> PeerId:
        return PeerId( id.bare | (BareId(id.kShift) << 48) )

    @staticmethod
    def FromSerialized(serialized : int) -> PeerId:
        flag = UserId.kReservedBit << 48
        legacy = not (serialized & (UserId.kReservedBit << 48))

        if (not legacy):
            return PeerId(serialized & (~flag))

        PeerIdMask = int(0xFFFFFFFF)
        PeerIdTypeMask = 0xF00000000
        PeerIdUserShift = 0x000000000
        PeerIdChatShift = 0x100000000
        PeerIdChannelShift = 0x200000000
        PeerIdFakeShift = 0xF00000000

        if ((serialized & PeerIdTypeMask) == PeerIdUserShift):
            return PeerId.FromChatIdType(UserId(BareId(serialized & PeerIdMask)))

        elif ((serialized & PeerIdTypeMask) == PeerIdChatShift):
            return PeerId.FromChatIdType(ChatId(BareId(serialized & PeerIdMask)))

        elif ((serialized & PeerIdTypeMask) == PeerIdChannelShift):
            return PeerId.FromChatIdType(ChannelId(BareId(serialized & PeerIdMask)))

        elif ((serialized & PeerIdTypeMask) == PeerIdFakeShift):
            return PeerId.FromChatIdType(FakeChatId(BareId(serialized & PeerIdMask)))
        
        return PeerId(0)

class FileKey(int):
    pass

class MetaDcId(type):
    
    def __new__(cls : Type[_T], clsName : str, bases : Tuple[type], attrs : Dict[str, Any]) -> _T:
        
        attrs['kDcShift'] = int(10000)
        attrs['Invalid'] = int(0)
        attrs['_0'] = int(0)
        attrs['_1'] = int(1)
        attrs['_2'] = int(2)
        attrs['_3'] = int(3)
        attrs['_4'] = int(4)
        attrs['_5'] = int(5)
        result = super().__new__(cls, clsName, bases, attrs)
        return result

class DcId(int):
    '''
    Data Center ID
    '''
    
    kDcShift : DcId = 10000 # type: ignore
    Invalid : DcId = 0 # type: ignore
    _0 : DcId = 0 # type: ignore
    _1 : DcId = 1 # type: ignore
    _2 : DcId = 2 # type: ignore
    _3 : DcId = 3 # type: ignore
    _4 : DcId = 4 # type: ignore
    _5 : DcId = 5 # type: ignore

    @staticmethod
    def BareDcId(shiftedDcId : Union[ShiftedDcId, DcId]) -> DcId:
        return DcId(shiftedDcId % DcId.kDcShift)


class ShiftedDcId(DcId):
    '''
    Shifted Data Center ID
    '''
    @staticmethod
    def ShiftDcId(dcId : DcId, value : int) -> ShiftedDcId:
        return ShiftedDcId(dcId + DcId.kDcShift * value)

class BuiltInDc(BaseObject): # type: ignore
    '''
    Default DC that is hard-coded
    '''

    def __init__(self, id : DcId, ip : str, port : int):
        self.id = id
        self.ip = ip
        self.port = port

class BuiltInDc(BuiltInDc):
        kBuiltInDcs = [
            BuiltInDc( DcId._1, "149.154.175.50" , 443 ),
            BuiltInDc( DcId._2, "149.154.167.51" , 443 ),
            BuiltInDc( DcId._2, "95.161.76.100"  , 443 ),
            BuiltInDc( DcId._3, "149.154.175.100", 443 ),
            BuiltInDc( DcId._4, "149.154.167.91" , 443 ),
            BuiltInDc( DcId._5, "149.154.171.5"  , 443 )]

        kBuiltInDcsIPv6 = [
            BuiltInDc( DcId._1, "2001:0b28:f23d:f001:0000:0000:0000:000a", 443 ),
            BuiltInDc( DcId._2, "2001:067c:04e8:f002:0000:0000:0000:000a", 443 ),
            BuiltInDc( DcId._3, "2001:0b28:f23d:f003:0000:0000:0000:000a", 443 ),
            BuiltInDc( DcId._4, "2001:067c:04e8:f004:0000:0000:0000:000a", 443 ),
            BuiltInDc( DcId._5, "2001:0b28:f23f:f005:0000:0000:0000:000a", 443 )]

        kBuiltInDcsTest = [
            BuiltInDc( DcId._1, "149.154.175.10" , 443 ),
            BuiltInDc( DcId._2, "149.154.167.40" , 443 ),
            BuiltInDc( DcId._3, "149.154.175.117", 443 )]

        kBuiltInDcsIPv6Test = [
            BuiltInDc( DcId._1, "2001:0b28:f23d:f001:0000:0000:0000:000e", 443 ),
            BuiltInDc( DcId._2, "2001:067c:04e8:f002:0000:0000:0000:000e", 443 ),
            BuiltInDc( DcId._3, "2001:0b28:f23d:f003:0000:0000:0000:000e", 443 )]

class dbi(int):
    Key = 0x00
    User = 0x01
    DcOptionOldOld = 0x02
    ChatSizeMaxOld = 0x03
    MutePeerOld = 0x04
    SendKeyOld = 0x05
    AutoStart = 0x06
    StartMinimized = 0x07
    SoundFlashBounceNotifyOld = 0x08
    WorkModeOld = 0x09
    SeenTrayTooltip = 0x0a
    DesktopNotifyOld = 0x0b
    AutoUpdate = 0x0c
    LastUpdateCheck = 0x0d
    WindowPositionOld = 0x0e
    ConnectionTypeOldOld = 0x0f
    # 0x10 reserved
    DefaultAttach = 0x11
    CatsAndDogsOld = 0x12
    ReplaceEmojiOld = 0x13
    AskDownloadPathOld = 0x14
    DownloadPathOldOld = 0x15
    ScaleOld = 0x16
    EmojiTabOld = 0x17
    RecentEmojiOldOldOld = 0x18
    LoggedPhoneNumberOld = 0x19
    MutedPeersOld = 0x1a
    # 0x1b reserved
    NotifyViewOld = 0x1c
    SendToMenu = 0x1d
    CompressPastedImageOld = 0x1e
    LangOld = 0x1f
    LangFileOld = 0x20
    TileBackgroundOld = 0x21
    AutoLockOld = 0x22
    DialogLastPath = 0x23
    RecentEmojiOldOld = 0x24
    EmojiVariantsOldOld = 0x25
    RecentStickers = 0x26
    DcOptionOld = 0x27
    TryIPv6Old = 0x28
    SongVolumeOld = 0x29
    WindowsNotificationsOld = 0x30
    IncludeMutedOld = 0x31
    MegagroupSizeMaxOld = 0x32
    DownloadPathOld = 0x33
    AutoDownloadOld = 0x34
    SavedGifsLimitOld = 0x35
    ShowingSavedGifsOld = 0x36
    AutoPlayOld = 0x37
    AdaptiveForWideOld = 0x38
    HiddenPinnedMessagesOld = 0x39
    RecentEmojiOld = 0x3a
    EmojiVariantsOld = 0x3b
    DialogsModeOld = 0x40
    ModerateModeOld = 0x41
    VideoVolumeOld = 0x42
    StickersRecentLimitOld = 0x43
    NativeNotificationsOld = 0x44
    NotificationsCountOld = 0x45
    NotificationsCornerOld = 0x46
    ThemeKeyOld = 0x47
    DialogsWidthRatioOld = 0x48
    UseExternalVideoPlayer = 0x49
    DcOptionsOld = 0x4a
    MtpAuthorization = 0x4b
    LastSeenWarningSeenOld = 0x4c
    SessionSettings = 0x4d
    LangPackKey = 0x4e
    ConnectionTypeOld = 0x4f
    StickersFavedLimitOld = 0x50
    SuggestStickersByEmojiOld = 0x51
    SuggestEmojiOld = 0x52
    TxtDomainStringOldOld = 0x53
    ThemeKey = 0x54
    TileBackground = 0x55
    CacheSettingsOld = 0x56
    AnimationsDisabled = 0x57
    ScalePercent = 0x58
    PlaybackSpeedOld = 0x59
    LanguagesKey = 0x5a
    CallSettingsOld = 0x5b
    CacheSettings = 0x5c
    TxtDomainStringOld = 0x5d
    ApplicationSettings = 0x5e
    DialogsFiltersOld = 0x5f
    FallbackProductionConfig = 0x60
    BackgroundKey = 0x61

    EncryptedWithSalt = 333
    Encrypted = 444

    Version = 666

class lskType(int):
    lskUserMap = 0x00
    lskDraft = 0x01 # data: PeerId peer
    lskDraftPosition = 0x02 # data: PeerId peer
    lskLegacyImages = 0x03 # legacy
    lskLocations = 0x04 # no data
    lskLegacyStickerImages = 0x05 # legacy
    lskLegacyAudios = 0x06 # legacy
    lskRecentStickersOld = 0x07 # no data
    lskBackgroundOldOld = 0x08 # no data
    lskUserSettings = 0x09 # no data
    lskRecentHashtagsAndBots = 0x0a # no data
    lskStickersOld = 0x0b # no data
    lskSavedPeersOld = 0x0c # no data
    lskReportSpamStatusesOld = 0x0d # no data
    lskSavedGifsOld = 0x0e # no data
    lskSavedGifs = 0x0f # no data
    lskStickersKeys = 0x10 # no data
    lskTrustedBots = 0x11 # no data
    lskFavedStickers = 0x12 # no data
    lskExportSettings = 0x13 # no data
    lskBackgroundOld = 0x14 # no data
    lskSelfSerialized = 0x15 # serialized self
    lskMasksKeys = 0x16 # no data

class BotTrustFlag(int):
    NoOpenGame = (1 << 0)
    Payment = (1 << 1)


class QByteArray(QByteArray):
    def dump(self):
            
        print(f"  ---- hexdump ----: [0x{hex(self.size())}:{self.size()}] ({id(self)})")

        size = self.size()
        data = self.data()
        out = ""
        asc = ""

        for i in range(size):

            if i % 16 == 0:
                if i != 0:
                    out += " |  "
                    out += asc + "\n"
                    asc = ""

                out += "  %0.4x " % i
            
            out += " %0.2x" % data[i]

            if ((i > 0) and ((i % 8) == 7) and ((i % 16) != 15)):
                out += " "

            if (data[i] < 0x20) or (data[i] > 0x7e):
                asc += "."
            else:
                asc += chr(data[i])

        i = size - 1
        if (((i % 16) != 0) and ((i % 16) == (i % 8))):
            out += " "
        
        while((i % 16) != 0):
            out += "   "
            i += 1
        
        out += " |  "
        print(out)

@staticmethod
def hexdump(byte : bytes):


    if isinstance(byte, QByteArray):
        byte = byte.data()
        
    print(f"  ---- hexdump ----: [0x{hex(byte.__len__())}:{byte.__len__()}] ({id(byte)})")

    size = byte.__len__()
    data = byte
    out = ""
    asc = ""

    for i in range(size):

        if i % 16 == 0:
            if i != 0:
                out += " |  "
                out += asc + "\n"
                asc = ""

            out += "  %0.4x " % i
        
        out += " %0.2x" % data[i]

        if ((i > 0) and ((i % 8) == 7) and ((i % 16) != 15)):
            out += " "

        if (data[i] < 0x20) or (data[i] > 0x7e):
            asc += "."
        else:
            asc += chr(data[i])

    i = size - 1
    if (((i % 16) != 0) and ((i % 16) == (i % 8))):
        out += " "
    
    while((i % 16) != 0):
        out += "   "
        i += 1
    
    out += " |  "
    print(out)



from __future__ import annotations
from .configs import *
from . import shared as td

import hashlib
import os
import tgcrypto

# if TYPE_CHECKING:
#     from ..opentele import *


class Serialize(BaseObject):

    @staticmethod
    def bytearraySize(arr : QByteArray):
        return sizeof(uint32) + arr.size()

    @staticmethod
    def bytesSize(arr : bytes):
        return sizeof(uint32) + len(arr)

    @staticmethod
    def stringSize(arr : str):
        return sizeof(uint32) + len(arr) + sizeof(ushort)
        
class Storage(BaseObject):

    class ReadSettingsContext(BaseObject):
        def __init__(self) -> None:
            
            self.fallbackConfigLegacyDcOptions : td.MTP.DcOptions = td.MTP.DcOptions(td.MTP.Environment.Production)
            self.fallbackConfigLegacyChatSizeMax = 0
            self.fallbackConfigLegacySavedGifsLimit = 0
            self.fallbackConfigLegacyStickersRecentLimit = 0
            self.fallbackConfigLegacyStickersFavedLimit = 0
            self.fallbackConfigLegacyMegagroupSizeMax = 0
            self.fallbackConfigLegacyTxtDomainString = str()
            self.fallbackConfig = QByteArray()

            self.cacheTotalSizeLimit = 0
            self.cacheTotalTimeLimit = 0
            self.cacheBigFileTotalSizeLimit = 0
            self.cacheBigFileTotalTimeLimit = 0

            self.themeKeyLegacy = FileKey(0)
            self.themeKeyDay = FileKey(0)
            self.themeKeyNight = FileKey(0)
            self.backgroundKeyDay = FileKey(0)
            self.backgroundKeyNight = FileKey(0)

            self.backgroundKeysRead = False
            self.tileDay = False
            self.tileNight = True
            self.tileRead = False

            self.langPackKey = FileKey(0)
            self.languagesKey = FileKey(0)

            self.mtpAuthorization = QByteArray()
            self.mtpLegacyKeys : typing.List[td.AuthKey] = []

            self.mtpLegacyMainDcId = 0 
            self.mtpLegacyUserId = 0
            self.legacyRead = False


    class FileReadDescriptor(BaseObject):
        def __init__(self) -> None:
            self.__version = 0
            self.__data = QByteArray()
            self.__buffer = QBuffer()
            self.__stream = QDataStream()

        @property
        def data(self) -> QByteArray:
            return self.__data
        @data.setter
        def data(self, value):
            self.__data = value

        @property
        def version(self) -> int:
            return self.__version
        @version.setter
        def version(self, value):
            self.__version = value

        @property
        def buffer(self) -> QBuffer:
            return self.__buffer

        @property
        def stream(self) -> QDataStream:
            return self.__stream

    class EncryptedDescriptor(BaseObject):
        def __init__(self, size : int = None) -> None:
            self.__data = QByteArray()
            self.__buffer = QBuffer()
            self.__stream = QDataStream()

            if size:
                fullSize = 4 + size
                if (fullSize & 0x0F):
                    fullSize += 0x10 - (fullSize & 0x0F)
                
                self.__data.reserve(fullSize)
                self.__data.resize(4)
                self.__buffer.setBuffer(self.__data)
                self.__buffer.open(QIODevice.OpenModeFlag.WriteOnly)
                self.__buffer.seek(4)
                self.__stream.setDevice(self.__buffer)
                self.__stream.setVersion(QDataStream.Version.Qt_5_1)

        def __del__(self):
            pass

        def finish(self) -> None:
            if self.__stream.device(): self.__stream.setDevice(None) # type: ignore
            if self.__buffer.isOpen(): self.__buffer.close()
            self.__buffer.setBuffer(None) # type: ignore

        @property
        def data(self) -> QByteArray:
            return self.__data
        @data.setter
        def data(self, value):
            self.__data = value

        @property
        def buffer(self) -> QBuffer:
            return self.__buffer

        @property
        def stream(self) -> QDataStream:
            return self.__stream

    class FileWriteDescriptor(BaseObject):
        def __init__(self, fileName : str, basePath : str, sync : bool = False) -> None:
            self.fileName = fileName
            self.basePath = basePath
            self.sync = sync
            self.init(fileName)

        def init(self, fileName : str):
            self.base = self.basePath + fileName
            self.safeData = QByteArray()
            self.buffer = QBuffer()
            self.buffer.setBuffer(self.safeData)

            result = self.buffer.open(QIODevice.OpenModeFlag.WriteOnly)
            Expects(result, "Can not open buffer, something went wrong")

            self.stream = QDataStream()
            self.stream.setDevice(self.buffer)

            self.md5 = bytes()
            self.fullSize = 0

        def writeData(self, data : QByteArray):
            Expects(self.stream.device() != None, "stream.device is missing")

            self.stream << data
            len = 0xffffffff if data.isNull() else data.size()

            if QSysInfo.Endian.ByteOrder != QSysInfo.Endian.BigEndian:
                def qbswap(source : int) -> int:
                    return 0 \
                        | ((source & 0x000000ff) << 24) \
                        | ((source & 0x0000ff00) << 8)  \
                        | ((source & 0x00ff0000) >> 8)  \
                        | ((source & 0xff000000) >> 24) \

                len = qbswap(len)
            
            self.md5 += len.to_bytes(4, 'little')
            self.md5 += data.data()
            self.fullSize += 4 + data.size()

        def writeEncrypted(self, data : Storage.EncryptedDescriptor, key : td.AuthKey):
            self.writeData(Storage.PrepareEncrypted(data, key))

        def finish(self):
            if not self.stream.device(): return
            self.stream.setDevice(None) # type: ignore
            
            self.md5 += self.fullSize.to_bytes(4, 'little')
            self.md5 += APP_VERSION.to_bytes(4, 'little')
            self.md5 += TDF_MAGIC
            self.md5 = hashlib.md5(self.md5).digest()

            self.buffer.close()
            
            finalData : bytes = self.safeData.data() + self.md5
            Storage.WriteFile(self.fileName, self.basePath, finalData)

        


    @staticmethod
    def PrepareEncrypted(data : EncryptedDescriptor, key : td.AuthKey) -> QByteArray:
        data.finish()

        toEncrypt = QByteArray(data.data)
        
        # prepare for encryption
        size = toEncrypt.size()
        fullSize = size
        if (fullSize & 0x0F):
            fullSize += 0x10 - (fullSize & 0x0F)
            toEncrypt.resize(fullSize)
            # TO BE ADDED
            # base::RandomFill(toEncrypt.data() + size, fullSize - size);
        
        # *(uint32*)toEncrypt.data() = size;
        toEncrypt = QByteArray(size.to_bytes(4, 'little')) + toEncrypt[4:]

        hashData = hashlib.sha1(toEncrypt)
        encrypted = QByteArray(hashData.digest())

        encrypted.resize(0x10 + fullSize)  # 128bit of sha1 - key128, sizeof(data), data
        
        encrypted = encrypted[:0x10] + Storage.aesEncryptLocal(toEncrypt, key, encrypted)

        return encrypted

    @staticmethod
    def WriteFile(fileName : str, basePath : str, data : bytes):
        
        dir = QDir(basePath)
        if not dir.exists():
            dir.mkpath(basePath)
            
        file = QFile(basePath + fileName + 's')
        if file.open(QIODevice.OpenModeFlag.WriteOnly):
            file.write(TDF_MAGIC)
            file.write(APP_VERSION.to_bytes(4, 'little'))
            file.write(data)
            file.close()
            return True

        raise OpenTeleException("what")
     
    @staticmethod   
    def ReadFile(fileName : str, basePath : str) -> FileReadDescriptor:

        to_try = ['s', '1', '0']
        tries_exception = None
        for chr in to_try:
            try:

                file = QFile(basePath + fileName + chr)
                if (not file.open(QIODevice.OpenModeFlag.ReadOnly)):
                    tries_exception = TFileNotFound(f"Could not open {fileName}")
                    continue

                
                magic = file.read(4)
                
                if magic != TDF_MAGIC:
                    tries_exception = TDataInvalidMagic("Invalid magic {magic} in file {fileName}")
                    file.close()
                    continue
                
                version = int.from_bytes(file.read(4), 'little')

                bytesdata = QByteArray(file.read(file.size()))

                dataSize = bytesdata.size() - 16

                check_md5 = bytesdata.data()[:dataSize]
                check_md5 += int(dataSize).to_bytes(4, 'little')
                check_md5 += int(version).to_bytes(4, 'little')
                check_md5 += magic
                check_md5 = hashlib.md5(check_md5)
                
                md5 = bytesdata.data()[dataSize:]
                
                if (check_md5.hexdigest() != md5.hex()):
                    tries_exception = TDataInvalidCheckSum("Invalid checksum {check_md5.hexdigest()} in file {fileName}")
                    file.close()
                    continue


                result = Storage.FileReadDescriptor()
                result.version = version

                bytesdata.resize(dataSize)
                result.data = bytesdata
                bytesdata = QByteArray()

                result.buffer.setBuffer(result.data)

                result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
                result.stream.setDevice(result.buffer)
                result.stream.setVersion(QDataStream.Version.Qt_5_1)

                file.close()
                return result
            except IOError as e:
                pass
        
        raise tries_exception if tries_exception else TFileNotFound("Could not open {fileName}")

    @staticmethod
    def ReadEncryptedFile(fileName : str, basePath : str, authKey : td.AuthKey) -> FileReadDescriptor:

        result = Storage.ReadFile(fileName, basePath)
        encrypted = QByteArray()
        result.stream >> encrypted

        try:
            data = Storage.DecryptLocal(encrypted, authKey)
        except OpenTeleException as e:
            result.stream.setDevice(None) # type: ignore
            if result.buffer.isOpen(): result.buffer.close()
            result.buffer.setBuffer(None) # type: ignore
            result.data = QByteArray()
            result.version = 0
            raise e
        
        result.stream.setDevice(None) # type: ignore
        if result.buffer.isOpen(): result.buffer.close()
        result.buffer.setBuffer(None) # type: ignore
        result.data = data.data
        
        result.buffer.setBuffer(result.data)
        result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
        result.buffer.seek(data.buffer.pos())
        result.stream.setDevice(result.buffer)
        result.stream.setVersion(QDataStream.Version.Qt_5_1)

        return result

    @staticmethod
    def ReadSetting(blockId : int, stream : QDataStream, version : int, context : ReadSettingsContext) -> bool:

        if (blockId == dbi.DcOptionOldOld):
            dcId = DcId(stream.readUInt32())
            host = stream.readQString()
            ip = stream.readQString()
            port = stream.readUInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyDcOptions.constructAddOne(
                dcId, td.MTP.DcOptions.Flag(0) , ip, port, bytes()
            )
            context.legacyRead = True

        elif (blockId == dbi.DcOptionOld):
            dcIdWithShift = ShiftedDcId(stream.readUInt32())
            flags = td.MTP.DcOptions.Flag(stream.readInt32())
            ip = stream.readQString()
            port = stream.readUInt32()
            
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyDcOptions.constructAddOne(
                dcIdWithShift, flags, ip, port, bytes()
            )
            context.legacyRead = True

        elif (blockId == dbi.DcOptionsOld):
            serialized = QByteArray()
            stream >> serialized
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyDcOptions.constructFromSerialized(
                serialized
            )
            context.legacyRead = True

        elif (blockId == dbi.ApplicationSettings):
            serialized = QByteArray()
            stream >> serialized
            ExpectStreamStatus(stream)

            # TO BE ADDED
        
        elif (blockId == dbi.ChatSizeMaxOld):
            maxSize = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyChatSizeMax = maxSize
            context.legacyRead = True

        elif (blockId == dbi.SavedGifsLimitOld):
            limit = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacySavedGifsLimit = limit
            context.legacyRead = True 
            
        elif (blockId == dbi.StickersRecentLimitOld):
            limit = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyStickersRecentLimit = limit
            context.legacyRead = True

        elif (blockId == dbi.StickersFavedLimitOld):
            limit = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyStickersFavedLimit = limit
            context.legacyRead = True

        elif (blockId == dbi.MegagroupSizeMaxOld):
            maxSize = stream.readInt32()
            ExpectStreamStatus(stream)

            context.fallbackConfigLegacyMegagroupSizeMax = maxSize
            context.legacyRead = True

        elif (blockId == dbi.User):
            userId = stream.readInt32()
            dcId = stream.readUInt32()
            ExpectStreamStatus(stream)

            context.mtpLegacyMainDcId = dcId
            context.mtpLegacyUserId = userId

        elif (blockId == dbi.Key):
            dcId = DcId(stream.readInt32())
            key = stream.readRawData(256)
            ExpectStreamStatus(stream)

            context.mtpLegacyKeys.append(
                td.AuthKey(key, td.AuthKeyType.ReadFromFile, dcId)
            )

        elif (blockId == dbi.MtpAuthorization):
            serialized = QByteArray()
            stream >> serialized
            ExpectStreamStatus(stream)

            context.mtpAuthorization = serialized

        return True

    @staticmethod
    def CreateLocalKey(salt : QByteArray, passcode: QByteArray = QByteArray()) -> td.AuthKey:
        hashKey = hashlib.sha512(salt)
        hashKey.update(passcode)
        hashKey.update(salt)

        iterationsCount = 1 if passcode.isEmpty() else 100000
        return td.AuthKey(hashlib.pbkdf2_hmac("sha512", hashKey.digest(), salt, iterationsCount, 256))

    @staticmethod
    def CreateLegacyLocalKey(salt : QByteArray, passcode: QByteArray = QByteArray()) -> td.AuthKey:

        iterationsCount = 1 if passcode.isEmpty() else 100000
        return td.AuthKey(hashlib.pbkdf2_hmac("sha512", passcode.data(), salt.data(), iterationsCount, 256))

    @staticmethod
    def aesEncryptLocal(src : QByteArray, authKey : td.AuthKey, key128 : QByteArray) -> QByteArray:
        aesKey, aesIv = authKey.prepareAES_oldmtp(key128, False)
        encrypted = tgcrypto.ige256_encrypt(src, aesKey, aesIv)
        return QByteArray(encrypted)
     
    @staticmethod   
    def aesDecryptLocal(src : QByteArray, authKey : td.AuthKey, key128 : QByteArray) -> QByteArray:
        aesKey, aesIv = authKey.prepareAES_oldmtp(key128, False)
        decrypted = tgcrypto.ige256_decrypt(src, aesKey, aesIv)
        return QByteArray(decrypted)

    @staticmethod
    def DecryptLocal(encrypted : QByteArray, authKey : td.AuthKey) -> EncryptedDescriptor:

        encryptedSize = encrypted.size()
        if (encryptedSize <= 16) or (encryptedSize & 0x0F):
            raise TDataBadEncryptedDataSize("Bad encrypted part size: {encryptedSize}")

        fullLen = encryptedSize - 16
        encryptedKey = encrypted[:16]
        encryptedData = encrypted[16:]

        decrypted = Storage.aesDecryptLocal(encryptedData, authKey, encryptedKey)
        checkHash = hashlib.sha1(decrypted).digest()[:16]
        if (checkHash != encryptedKey):
            raise TDataBadDecryptKey("Bad decrypt key, data not decrypted - incorrect password?")

        dataLen = int.from_bytes(decrypted[:4], 'little') # *(const uint32*)decrypted.constData();
        if (dataLen > decrypted.size()) or (dataLen <= fullLen - 16) or (dataLen < 4):
            raise TDataBadDecryptedDataSize("Bad decrypted part size: {encryptedSize}, fullLen: {fullLen}, decrypted size: {decrypted.__len__()}")

        decrypted.resize(dataLen)
        result = Storage.EncryptedDescriptor()
        result.data = decrypted

        decrypted = QByteArray()
        
        result.buffer.setBuffer(result.data)
        result.buffer.open(QIODevice.OpenModeFlag.ReadOnly)
        result.buffer.seek(4) # skip len
        result.stream.setDevice(result.buffer)
        result.stream.setVersion(QDataStream.Version.Qt_5_1)

        return result

    @staticmethod
    def ComposeDataString(dataName : str, index : int):
        result = dataName
        result.replace("#", "")
        if (index > 0): result += f"#{index + 1}"
        return result

    @staticmethod
    def ComputeDataNameKey(dataName : str) -> int:
        md5 = hashlib.md5(dataName.encode("utf-8")).digest()
        return int.from_bytes(md5, 'little')

    @staticmethod
    def ToFilePart(val : int):
        result = str()
        for i in range(0, 0x10):
            v = val & 0xF
            if v < 0x0A:
                result += chr(ord('0') + v)
            else:
                result += chr(ord('A') + (v - 0x0A))
            val >>= 4
        return result

    @staticmethod
    def RandomGenerate(size : int) -> QByteArray:
        return QByteArray(os.urandom(size))

    @staticmethod
    def GetAbsolutePath(path : str = None) -> str:

        if path == None or path == "":
            path = os.getcwd()
        
        path = os.path.abspath(path) + "\\"
        return path




from __future__ import annotations
from .shared import *


# class TDesktop(TDesktop):

#     def ImportTelethonAccount(self, telethonClient : telethon.TelegramClient) -> bool:

#         excpt.Expects(self.accountsCount < TDesktop.kMaxAccounts,
#         exception=excpt.MaxAccountLimit("You can't have more than 3 accounts in one TDesktop clent.\n"\
#                                         "Please create another instance of TDesktop or use Account.FromTelethon() to create an Account() independently"))
        
        
        
#         # newAccount.serializeMtpAuthorization()
#         # newAccount.prepareToStart(self.__localKey)
#         # if not newAccount.isLoaded():
#         #     return False 
        
#         # self.accounts.append(newAccount)
#         return True

# class TDesktop(TDesktop):
#     async def ToNewTelethonSession(self, session_name : str, customAPI : API = None) -> TelegramClient:
        
#         resolve_file_name = session_name
#         if resolve_file_name[-8:] != ".session": resolve_file_name += ".session"

#         if os.path.exists(resolve_file_name):
#             os.remove(resolve_file_name)

#         if customAPI != None:
#             newClient = TelegramClient(session_name, customAPI.api_id, customAPI.api_hash, device_model=customAPI.pid) # pass hook data pid through device_model
#         else:
#             defaultAPI =  API.TelegramDesktop
#             newClient = TelegramClient(session_name, defaultAPI.api_id, defaultAPI.api_hash, )
        
#         oldClient = self.ToTelethon()

#         await oldClient.connect()
#         await newClient.connect()
#         oldMe = await oldClient.get_me()
        
#         qr_login = await newClient.qr_login()
        
#         try:
#             resp = await oldClient(tl.functions.auth.AcceptLoginTokenRequest(qr_login.token))
#         except BaseException as e:
#             pass

#         try:
#             await qr_login.wait()
#         except telethon.errors.SessionPasswordNeededError as e:

#             # two-step verification
#             username = oldMe.username if (oldMe.username != None) else utils.get_display_name(oldMe)
#             password = input(f"Password for {username}: ")

#             pwd = await newClient(tl.functions.account.GetPasswordRequest())
#             result = await newClient(tl.functions.auth.CheckPasswordRequest(
#                 pwd_mod.compute_check(pwd, password)))
        
#             newClient._on_login(result.user)

#         await newClient.get_me()
#         return newClient



import pathlib
from setuptools import find_packages, setup
import re

README = (pathlib.Path(__file__).parent / "README.md").read_text()

PACKAGE_NAME        = "opentele"
VERSION             = "1.1.2"
SOURCE_DIRECTORY    = "src"

with open("requirements.txt") as data:
    requirements = [
        line for line in data.read().split("\n")
        if line and not line.startswith("#")
    ]

setup(

    name=PACKAGE_NAME,
    version=VERSION,
    license="MIT",

    description="A python library created to make life easier for Telegram API Developers.",
    long_description=README,
    long_description_content_type="text/markdown",
    
    url="https://github.com/thedemons/opentele",
    author="thedemons",
    author_email="thedemons@gmail.com",
    
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 5 - Production/Stable",
    ],
    
    keywords=[
        "tdata",
        "tdesktop",
        "telegram",
        "telethon",
        "opentele",
    ],
    
    packages=[PACKAGE_NAME] + find_packages(PACKAGE_NAME),
    package_dir={PACKAGE_NAME: SOURCE_DIRECTORY},
    include_package_data=True,
    install_requires=requirements,
)



from __future__ import annotations
from .configs import *
from . import shared as td

import hashlib

# if TYPE_CHECKING:
#     from ..opentele import *

class AuthKeyType(IntEnum):
    """
    Type of `AuthKey`
    
    ### Attributes:
        Generated (`IntEnum`):
            Generated key

        Temporary (`IntEnum`):
            Temporary key

        ReadFromFile (`IntEnum`):
            Key red from file

        Local (`IntEnum`):
            Local key
    """
    Generated       = 0
    Temporary       = 1
    ReadFromFile    = 2
    Local           = 3

class AuthKey(BaseObject):
    """
    Authorization key used for [MTProto](https://core.telegram.org/mtproto)
    It's also used to encrypt and decrypt local tdata

    ### Attributes:
        DcId (DcId):
            Data Center ID (from 1 to 5).
            
        type (AuthKeyType):
            Type of the key.

        key (bytes):
            The actual key, 256 `bytes` in length.

    """

    kSize = 256
    def __init__(self, key : bytes = bytes(), type : AuthKeyType = AuthKeyType.Generated, dcId : DcId = DcId.Invalid) -> None: # type: ignore
        self.__type = type
        self.__dcId = dcId
        self.__key = key
        # if (type == self.Type.Generated) or (type == self.Type.Temporary):
            # self.__creationtime = ...
        self.__countKeyId()

    @property
    def dcId(self) -> DcId:
        return self.__dcId

    @property
    def type(self) -> AuthKeyType:
        return self.__type

    @property
    def key(self) -> bytes:
        return self.__key
    
    def write(self, to : QDataStream) -> None:
        to.writeRawData(self.key)

    def __countKeyId(self) -> None:
        hash = hashlib.sha1(self.__key).digest()
        self.__keyId = int.from_bytes(hash[12:12+8], 'little')
    
    def prepareAES_oldmtp(self, msgKey : bytes, send : bool) -> typing.Tuple[bytes, bytes]:
        x = 0 if send else 8
        sha1_a = hashlib.sha1(msgKey[:16] + self.__key[x : x+32]).digest()
        sha1_b = hashlib.sha1(self.__key[x+32 : x+32+16] + msgKey[:16] + self.__key[x+48 : x+48+16]).digest()
        sha1_c = hashlib.sha1(self.__key[x+64 : x+64+32] + msgKey[:16]).digest()
        sha1_d = hashlib.sha1(msgKey[:16] + self.__key[x+96 : x+96+32]).digest()

        aesKey = sha1_a[:8] + sha1_b[8:8+12] + sha1_c[4:4+12]
        aesIv = sha1_a[8:8+12] + sha1_b[:8] + sha1_c[16:16+4] + sha1_d[:8]

        return aesKey, aesIv

    @staticmethod
    def FromStream(stream : QDataStream, type : AuthKeyType = AuthKeyType.ReadFromFile, dcId : DcId = DcId(0)) -> AuthKey:
        keyData = stream.readRawData(AuthKey.kSize)
        return AuthKey(keyData, type, dcId)